/*                
 *                        Licensing Notification
 *
 * Permission to use, reproduce, prepare derivative works, and to redistribute
 * to others this software and its documentation is hereby granted, provided
 * that this notice is retained thereon and on all copies or modifications. UChicago
 * Argonne, LLC, Tech-X Corporation, and all other contributors make no representations
 * as to the suitability and operability of this software for any purpose. It is
 * provided "as is" without express or implied warranty.
 *
 */

#ifndef CSR_TRI_SOLVE_PART_CPU_H
#define CSR_TRI_SOLVE_PART_CPU_H

// -----------------------------------------------------------------------------

#include "uplo_type.h"
#include "csr_matrix_data.h"
#include <cstddef>
//#include <iostream>

// -----------------------------------------------------------------------------

const unsigned int CSR_TRI_SOLVE_PART_CPU_NUM_VORDER = 2;

// -----------------------------------------------------------------------------

// solves the lower/upper triangular block given in [block_start,block_end)
// only. A, b, x reside in CPU memory. "A" points to the "full" data and
// not just to that block, but "x" and "y" point to the necessary data
// only.  That is, "x" and "y" point to a block of size [0, block_end - block_start).
template<typename index_type, typename value_type>
bool csr_tri_solve_part_cpu(const csr_matrix_data<index_type, value_type>& A,
			    index_type* nnzPerRowDiagBlock,
                            index_type block_start,
                            index_type block_end,
                            uplo_enum uplo_type,
                            const value_type* b,
                            value_type* x,
                            index_type num_vecs,
                            index_type next_vec_offset,
                            unsigned int vorder);

// Internal APIs.  These don't do any error checking.
template<typename index_type, typename value_type>
void csr_tri_solve_part_up(const csr_matrix_data<index_type, value_type>& A,
			   index_type* nnzPerRowDiagBlock,
                           index_type block_start,
                           index_type block_end,
                           const value_type* b,
                           value_type* x,
                           index_type num_vecs,
                           index_type next_vec_offset,
                           unsigned int vorder);

template<typename index_type, typename value_type>
void csr_tri_solve_part_lo(const csr_matrix_data<index_type, value_type>& A,
			   index_type* nnzPerRowDiagBlock,
                           index_type block_start,
                           index_type block_end,
                           const value_type* b,
                           value_type* x,
                           index_type num_vecs,
                           index_type next_vec_offset,
                           unsigned int vorder);

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

template<typename index_type, typename value_type>
bool csr_tri_solve_part_cpu(const csr_matrix_data<index_type, value_type>& A,
			    index_type* nnzPerRowDiagBlock,
                            index_type block_start,
                            index_type block_end,
                            uplo_enum uplo_type,
                            const value_type* b,
                            value_type* x,
                            index_type num_vecs,
                            index_type next_vec_offset,
                            unsigned int vorder)
{
  //TODO: we don't check whether
  //
  // 1. diagonal entries actually exist
  // 2. diagonal contains any zeros
  // 3. the matrix is really triangular
  //
  // We probably never will.
  
  bool success = false;
  
  if( A.num_rows == A.num_cols &&
	    A.num_rows > 0 &&
      A.row_offsets &&
      A.column_indices &&
      A.values &&
      0 <= block_start &&
      block_start < block_end &&
	    block_end <= A.num_rows &&
      b &&
      x &&
      num_vecs > 0 &&
      next_vec_offset >= block_end - block_start &&
      vorder < CSR_TRI_SOLVE_PART_CPU_NUM_VORDER)
	{
          if(uplo_type == uplo_up_only)
          {
            csr_tri_solve_part_up(A, nnzPerRowDiagBlock, block_start, block_end, b, x, num_vecs, next_vec_offset, vorder);
            success = true;
          }
          else if(uplo_type == uplo_lo_only)
          {
            csr_tri_solve_part_lo(A, nnzPerRowDiagBlock, block_start, block_end, b, x, num_vecs, next_vec_offset, vorder);
            success = true;
          }
	}
  
  return success;
}

// -----------------------------------------------------------------------------

template<typename index_type, typename value_type>
void csr_tri_solve_part_up(const csr_matrix_data<index_type, value_type>& A,
			   index_type* nnzPerRowDiagBlock,
                           index_type block_start,
                           index_type block_end,
                           const value_type* b,
                           value_type* x,
                           index_type num_vecs,
                           index_type next_vec_offset,
                           unsigned int vorder)
{
  const index_type* A_row_offsets    = A.row_offsets;
  const index_type* A_column_indices = A.column_indices;
  const value_type* A_values         = A.values;

  value_type sum;  
  value_type* v;
  index_type* vi;
  index_type i, nz;

  index_type row_id = block_end;
  index_type start = A_row_offsets[row_id--];
  v= const_cast<value_type *> (&(A_values[start-1]));
  vi= const_cast<index_type *> (&(A_column_indices[start-1]));
  v++; vi++;
  x-=block_start;
  b-=block_start;
  index_type end = A_row_offsets[block_end];
  for (i=block_end-1; i>=block_start; i--) {
    nz  =  end -  A_row_offsets[i];
    sum = b[i];
    v  -= nz; vi -= nz;
    index_type nnz = nnzPerRowDiagBlock[i];
    value_type diag = v[0];
#if 1
    for(index_type __i=1; __i<nnz; __i++)
      sum -= v[__i] * x[vi[__i]];
#else
    {
      v++; vi++;
      index_type nnzTemp = nnz-1;
      while (nnzTemp >= 4) {
        sum -=  (v[0] * x[vi[0]] + v[1] * x[vi[1]] + v[2] * x[vi[2]] + v[3] * x[vi[3]]);
        v  += 4; vi += 4; nnzTemp -= 4;
      }
      switch (nnzTemp & 0x3) {
      case 3: sum -= *v++ * x[*vi++];
      case 2: sum -= *v++ * x[*vi++];
      case 1: sum -= *v++ * x[*vi++];
      }
      v-=nnz; vi-=nnz;
    }
#endif
    x[i] = sum/diag;
    end -= nz;
  }	  
  x+=block_start;
  b+=block_start;  
}

// -----------------------------------------------------------------------------

template<typename index_type, typename value_type>
void csr_tri_solve_part_lo(const csr_matrix_data<index_type, value_type>& A,
			   index_type* nnzPerRowDiagBlock,
                           index_type block_start,
                           index_type block_end,
                           const value_type* b,
                           value_type* x,
                           index_type num_vecs,
                           index_type next_vec_offset,
                           unsigned int vorder)
{
  const index_type* A_row_offsets    = A.row_offsets;
  const index_type* A_column_indices = A.column_indices;
  const value_type* A_values         = A.values;

  value_type sum;
  value_type* v;
  index_type* vi;
  index_type i, nz;

  index_type start = A_row_offsets[block_start];
  v= const_cast<value_type *> (&(A_values[start]));
  vi= const_cast<index_type *> (&(A_column_indices[start]));

  b-=block_start;
  x-=block_start;
  index_type nnz, nzMinusDiag;
  for (i=block_start; i<block_end; i++) {
    nz  =  A_row_offsets[i+1] -  start;
    nnz = nnzPerRowDiagBlock[i];
    nzMinusDiag = nz-nnz;
    v+=nzMinusDiag; vi+=nzMinusDiag;
    sum = b[i];

#if 1
    for(index_type __i=0; __i<nnz-1; __i++) sum -= v[__i] * x[vi[__i]];
    v  += nnz; vi += nnz;
#else
    {
      index_type nnzTemp = nnz-1;
      while (nnzTemp >= 4) {
        sum -=  (v[0] * x[vi[0]] + v[1] * x[vi[1]] + v[2] * x[vi[2]] + v[3] * x[vi[3]]);
        v  += 4; vi += 4; nnzTemp -= 4;
      }
      switch (nnzTemp & 0x3) {
      case 3: sum -= *v++ * x[*vi++];
      case 2: sum -= *v++ * x[*vi++];
      case 1: sum -= *v++ * x[*vi++];
      }
      v++; vi++;
    }
#endif
    x[i] = sum;
    start += nz;
  }
  b +=block_start;
  x +=block_start;
}

// -----------------------------------------------------------------------------

#endif // CSR_TRI_SOLVE_PART_CPU_H
